"""Reverse Distillation Model."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .lightning_model import ReverseDistillation, ReverseDistillationLightning

__all__ = ["ReverseDistillation", "ReverseDistillationLightning"]
