"""Deep Feature Extraction (DFM) model."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .lightning_model import Dfm, DfmLightning

__all__ = ["Dfm", "DfmLightning"]
